/*
 *  Copyright (c) 2007-2008 Cyrille Berger <cberger@cberger.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * either version 2, or (at your option) any later version of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "Debug.h"

#include <map>
#include <fstream>

#include <llvm/Module.h>
#include <llvm/Value.h>
#include <llvm/Type.h>
#include <llvm/System/Path.h>
#include <llvm/Support/raw_ostream.h>

#include "Function.h"
#include "LLVMBackend/ExpressionResult_p.h"
#include "Macros_p.h"
#include "PixelDescription.h"
#include "Region.h"
#include "ScopedName.h"
#include "Token_p.h"
#include "Type.h"
#include "Type_p.h"
#include "Value.h"
#include "Parameter.h"
#include "Color.h"

using namespace GTLCore;

struct FunctionDebugInfo {
  FunctionDebugInfo() : enabled( true ) {}
  bool enabled;
};

struct FileDebugInfo {
  FileDebugInfo() : enabled( true ) {}
  bool enabled;
  std::map<GTLCore::String, FunctionDebugInfo> functionsDebugInfo;
};

struct LibraryDebugInfo {
  LibraryDebugInfo() : enabled( true ) {}
  bool enabled;
  std::map<GTLCore::String, FileDebugInfo> filesDebugInfo;
};

struct Debug::Private {
  Private();
  ~Private();
  // Static
  static Debug::Private* instance();
  static Debug::Private* s_instance;
  // Helper members
  static bool isEnabled( const std::map< GTLCore::String, LibraryDebugInfo >& informationRecords, const GTLCore::String& _libraryName, const GTLCore::String& _fileName, const GTLCore::String& _functionName );
  static String extractFunctionName( const String& );
  static void readConfigFile( const String& _fileName, std::map< GTLCore::String, LibraryDebugInfo >& destination);
  static std::ostream& report( std::ostream&, const std::map< GTLCore::String, LibraryDebugInfo >& _informationRecords, const String& _streamName, const String& _libraryName, const String& _fileName, int _line, const String& _functionName );
  // Fields
  std::ostream* m_debugStream;
  std::ostream* m_warningStream;
  std::ostream* m_errorStream;
  std::ostream* m_voidStream;
  std::map< GTLCore::String, LibraryDebugInfo > m_librariesDebugInfo;
  std::map< GTLCore::String, LibraryDebugInfo > m_librariesWarningInfo;
  std::map< GTLCore::String, LibraryDebugInfo > m_librariesErrorInfo;
};

Debug::Private* Debug::Private::s_instance = 0;

STATIC_DELETER( Debug )
{
  delete Debug::Private::s_instance;
}

class NullStream : public std::ostream {
  public:
  	NullStream() : std::ostream(0) {}
};

Debug::Private::Private()
{
  m_debugStream = &std::cerr;
  m_warningStream = &std::cerr;
  m_errorStream = &std::cerr;
  m_voidStream = new NullStream;
//   m_librariesDebugInfo[ "GTLCore" ].enabled = false;
  // Read ${HOME}/.OpenGTLDebugConfig
  readConfigFile( ".OpenGTLDebugConfig", m_librariesDebugInfo );
  readConfigFile( ".OpenGTLWarningConfig", m_librariesWarningInfo );
  readConfigFile( ".OpenGTLErrorConfig", m_librariesErrorInfo );
}

Debug::Private::~Private()
{
  delete m_voidStream;
}

void Debug::Private::readConfigFile( const String& _fileName, std::map< GTLCore::String, LibraryDebugInfo >& _destination)
{
  llvm::sys::Path path = llvm::sys::Path::GetUserHomeDirectory();
  path.appendComponent( (const std::string&)_fileName);
  if( path.exists() )
  {
    std::ifstream in;
    in.open(path.c_str() );
    if(in)
    {
      std::string cstr;
      std::getline(in,cstr);
      GTLCore::String str(cstr);
      while ( in ) {
        if( not str.isEmpty() and str[0] != '#' )
        {
          std::vector< String > splited = str.split( " =,", false );
          if( splited.size() >= 2 and splited.size() <= 4 )
          {
            std::vector< String >::iterator it = --splited.end();
            bool status = (*(--splited.end()) == "true");
            it = splited.begin();
            LibraryDebugInfo& ldi = _destination[ *it ];
            if( splited.size() == 2 )
            {
              ldi.enabled = status;
            } else {
              FileDebugInfo& fdi = ldi.filesDebugInfo[ *(++it) ];
              if( splited.size() == 3 )
              {
                fdi.enabled = status;
              } else {
                FunctionDebugInfo& fundi = fdi.functionsDebugInfo[ *(++it) ];
                fundi.enabled = status;
              }
            }
          }
        }
        std::getline(in,cstr);
        str = cstr;
      }
    }
  }
}


bool Debug::Private::isEnabled( const std::map< GTLCore::String, LibraryDebugInfo >& informationRecords, const GTLCore::String& _libraryName, const GTLCore::String& _fileName, const GTLCore::String& _functionName )
{
  std::map< GTLCore::String, LibraryDebugInfo >::const_iterator ldi = informationRecords.find(_libraryName );
  if( ldi == informationRecords.end() ) return true;
  if( not ldi->second.enabled ) return false;
  std::map< GTLCore::String, FileDebugInfo>::const_iterator fdi = ldi->second.filesDebugInfo.find( _fileName );
  if( fdi == ldi->second.filesDebugInfo.end() ) return true;
  if( not fdi->second.enabled ) return false;
  std::map< GTLCore::String, FunctionDebugInfo>::const_iterator fundi = fdi->second.functionsDebugInfo.find( _functionName );
  if( fundi == fdi->second.functionsDebugInfo.end() ) return true;
  return fundi->second.enabled;
}

String Debug::Private::extractFunctionName( const String& _str)
{
  int posBeg = 0;
  int posEnd = 0;
  for( size_t i = 0; i < _str.length(); ++i)
  {
    if( _str[i] == '(' )
    {
      posEnd = i;
      break;
    }
  }
  for( size_t i = posEnd; i > 0; --i)
  {
    if( _str[i] == ' ' )
    {
      posBeg = i + 1;
      break;
    }
  }
  return _str.substr( posBeg, posEnd - posBeg );
}

std::ostream& Debug::Private::report( std::ostream& _stream, const std::map< GTLCore::String, LibraryDebugInfo >& _informationRecords, const String& _streamName, const String& _libraryName, const String& _fileName, int _line, const String& _functionName )
{
  GTLCore::String fileName = llvm::sys::Path((const std::string&)_fileName).getLast().str();
  GTLCore::String functionName = Private::extractFunctionName( _functionName );
  if( isEnabled( _informationRecords, _libraryName, fileName, functionName ) )
  {
    _stream << _libraryName << " (" << _streamName << "): in " << fileName << " at " << _line << " in " << functionName << ": ";
    return _stream;
  } else {
    return *Private::instance()->m_voidStream;
  }
}

bool Debug::isDebugEnabled( const String& _libraryName, const String& _fileName, const String& _functionName )
{
  Private* p = Private::instance();
  return p->isEnabled( p->m_librariesDebugInfo, _libraryName, _fileName, _functionName );
}

std::ostream& Debug::debug( const GTLCore::String& _libraryName, const GTLCore::String& _fileName, int _line, const GTLCore::String& _functionName )
{
  Private* p = Private::instance();
  return Private::report( *p->m_debugStream, p->m_librariesDebugInfo, "Debug", _libraryName, _fileName, _line, _functionName );
}

std::ostream& Debug::warning( const GTLCore::String& _libraryName, const GTLCore::String& _fileName, int _line, const GTLCore::String& _functionName )
{
  Private* p = Private::instance();
  return Private::report( *p->m_warningStream, p->m_librariesWarningInfo, "Warning", _libraryName, _fileName, _line, _functionName );
}

std::ostream& Debug::error( const GTLCore::String& _libraryName, const GTLCore::String& _fileName, int _line, const GTLCore::String& _functionName )
{
  Private* p = Private::instance();
  return Private::report( *p->m_errorStream, p->m_librariesErrorInfo, "Error", _libraryName, _fileName, _line, _functionName );
}


Debug::Private* Debug::Private::instance()
{
  if( s_instance == 0) s_instance = new Debug::Private;
  return s_instance;
}


namespace LLVMBackend {
  std::ostream& operator<< (std::ostream& _ostr, const ExpressionResult& _expressionResult)
  {
    _ostr << "[value = " << *_expressionResult.value() << " type = " << *_expressionResult.type();
    return _ostr;
  }
}
  
namespace GTLCore {
  std::ostream& operator<< (std::ostream& _ostr, const PixelDescription& _pixelDescription)
  {
    _ostr << "[";
    if( _pixelDescription.hasSameTypeChannels() )
    {
      _ostr << _pixelDescription.channels() << " x " << *_pixelDescription.channelTypes()[0];
    } else {
      for( std::size_t i = 0; i < _pixelDescription.channels(); ++i)
      {
        _ostr << *_pixelDescription.channelTypes()[i];
        if( i != _pixelDescription.channels() - 1 )
        {
          _ostr << ", ";
        }
      }
    }
    _ostr << "]";
    return _ostr;
  }
  
  std::ostream& operator<< (std::ostream& ostr, const RegionI& region)
  {
    ostr << "[" << region.x() << "," << region.y() << "," << region.columns() << "," << region.rows() << "]" ;
    return ostr;
  }
  
  std::ostream& operator<< (std::ostream& ostr, const RegionF& region)
  {
    ostr << "[" << region.x() << "," << region.y() << "," << region.columns() << "," << region.rows() << "]" ;
    return ostr;
  }
  
  std::ostream& operator<< (std::ostream& ostr, const GTLCore::ScopedName& name)
  {
    return ostr <<name.toString() ;
  }

  std::ostream& operator<< (std::ostream& ostr, const GTLCore::Type& type)
  {
    switch(type.dataType())
    {
      case Type::UNDEFINED:
        ostr << "UNDEFINED";
        break;
      case Type::BOOLEAN:
        ostr <<"BOOLEAN";
        break;
      case Type::INTEGER8:
        ostr <<"INTEGER8";
        break;
      case Type::UNSIGNED_INTEGER8:
        ostr <<"UNSIGNED_INTEGER8";
        break;
      case Type::INTEGER16:
        ostr <<"INTEGER16";
        break;
      case Type::UNSIGNED_INTEGER16:
        ostr <<"UNSIGNED_INTEGER16";
        break;
      case Type::INTEGER32:
        ostr <<"INTEGER32";
        break;
      case Type::UNSIGNED_INTEGER32:
        ostr <<"UNSIGNED_INTEGER32";
        break;
      case Type::INTEGER64:
        ostr <<"INTEGER64";
        break;
      case Type::UNSIGNED_INTEGER64:
        ostr <<"UNSIGNED_INTEGER64";
        break;
      case Type::FLOAT64:
        ostr <<"FLOAT64";
        break;
      case Type::FLOAT32:
        ostr <<"FLOAT32";
        break;
      case Type::FLOAT16:
        ostr <<"FLOAT16";
        break;
      case Type::VOID:
        ostr <<"VOID";
        break;
      case Type::STRUCTURE:
        ostr <<"STRUCTURE";
        break;
      case Type::ARRAY:
        ostr <<"ARRAY[" << *type.embeddedType() << "]";
        break;
      case Type::POINTER:
        ostr <<"POINTER";
        break;
      case Type::VECTOR:
        ostr <<"VECTOR[" << type.vectorSize() << "," << *type.embeddedType() << "]";
        break;
      case Type::MATRIX:
        ostr <<"MATRIX[]";
    }
    return ostr;
  }
  std::ostream& operator<< (std::ostream& ostr, const GTLCore::Token& token)
  {
    ostr << Token::typeToString( token.type );
    if( token.isPrimary() )
    {
      ostr << " primary";
    }
    return ostr;
  }
  std::ostream& operator<< (std::ostream& ostr, const GTLCore::Color& color)
  {
    ostr << "rgba[" << color.red() << ";" << color.green() << ";" << color.blue() << ";" << color.alpha() << "]";
    return ostr;
  }
  std::ostream& operator<< (std::ostream& ostr, const Value& value)
  {
    ostr << *value.type() << ": ";
    switch( value.type()->dataType() )
    {
      case Type::BOOLEAN:
        ostr << value.asBoolean();
        break;
      case Type::FLOAT32:
        ostr << value.asFloat32();
        break;
      case Type::INTEGER32:
        ostr << value.asInt32();
        break;
      case Type::UNSIGNED_INTEGER32:
        ostr << value.asUnsignedInt32();
        break;
      case Type::ARRAY:
        ostr << "[ ";
        foreach( const Value& val, *value.asArray() )
        {
          ostr << val << ", ";
        }
        ostr << " ]";
        break;
      case Type::STRUCTURE:
        if (value.type() == Type::Color)
        {
          ostr << value.asColor();
        }
      default:
        ostr << " unknown";
    }
    return ostr;
  }
  std::ostream& operator<< (std::ostream& ostr, const Function& function)
  {
    ostr << function.name() << " : " << *function.returnType() << " ( ";
    
    for( std::size_t i = 0; i < function.parameters().size(); ++i )
    {
      ostr << *function.parameters()[i].type();
      if( i != function.parameters().size() - 1 )
      {
        ostr << ", ";
      }
    }
    
    ostr << " )";
    return ostr;
  }

}


#ifdef OPENGTL_ENABLE_DEBUG_OUTPUT

#include <llvm/DerivedTypes.h>
#include <llvm/Function.h>
#include <llvm/Type.h>

void compareFunctionParameters( llvm::Value* func, const std::vector<llvm::Value*>& params  )
{
  GTL_ASSERT(func);
  const llvm::FunctionType *FTy =
    llvm::cast<llvm::FunctionType>(llvm::cast<llvm::PointerType>(func->getType())->getElementType());
  
  GTL_DEBUG(  *FTy );
  
  GTL_ASSERT( params.size() == FTy->getNumParams() or
          (FTy->isVarArg() and params.size() > FTy->getNumParams()) );
  
  for (unsigned i = 0; i < params.size(); ++i) {
    if( i < FTy->getNumParams() and (FTy->getParamType(i) != params[i]->getType()) )
    {
      GTL_DEBUG( "Wrong parameter " << i << " : " << FTy->getParamType(i) << " => " << *FTy->getParamType(i) << " but got " << params[i] << " => " << *params[i]->getType() );
    } else {
      GTL_DEBUG( "Parameter " << i << " : " << FTy->getParamType(i) << " => " << *FTy->getParamType(i) << " and got " << params[i]->getType() << " => " << *params[i]->getType() );
    }
  }
}

#endif

namespace llvm {
#define MAKE_OPERATOR(_type_)                                             \
  std::ostream& operator<< (std::ostream& ostr, const llvm::_type_& val)  \
  {                                                                       \
    std::string str;                                                      \
    llvm::raw_string_ostream rs(str);                                     \
    rs << val;                                                            \
    ostr << str;                                                          \
    return ostr;                                                          \
  }
  MAKE_OPERATOR(Module)
  MAKE_OPERATOR(Type)
  MAKE_OPERATOR(Value)
}

