/*
 *  Copyright (c) 2008 Cyrille Berger <cberger@cberger.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * either version 2, or (at your option) any later version of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef _OPENSHIVA_CODE_GENERATOR_P_H_
#define _OPENSHIVA_CODE_GENERATOR_P_H_

#include <vector>
#include "Kernel_p.h"

namespace llvm {
  class BasicBlock;
  class Function;
  class Module;
  class Value;
}

namespace GTLCore {
  class Function;
  class ModuleData;
  class Type;
  class PixelDescription;
}

namespace LLVMBackend {
  class GenerationContext;
}

namespace OpenShiva {
  class Kernel;
  /**
   * @internal
   * Provide code generation function for specific features of OpenShiva
   * @ingroup OpenShiva
   */
  class CodeGenerator {
    public:
      /**
       * Copy a pixel from memory to the vector.
       */
      static llvm::BasicBlock* memToPixel( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _dataPointer, llvm::Value* _pixel, llvm::Value* _image );
      static llvm::BasicBlock* pixelToMem( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _pixel, llvm::Value* _dataPointer, llvm::Value* _image, llvm::Value* channelsFlags );
      static llvm::Function* generateEvaluatePixels( Kernel* _kernel, llvm::Module* _module, int _channels );
      static llvm::Function* generateMemToVec( GTLCore::ModuleData* _moduleData, llvm::Module* _module, const GTLCore::PixelDescription& _pixelDescription );
      static llvm::Function* generateVecToMem( GTLCore::ModuleData* _moduleData, llvm::Module* _module, const GTLCore::PixelDescription& _pixelDescription );
      /**
       * Generate a sample nearest function.
       */
      static llvm::Function* generateImageSampleNearest( GTLCore::ModuleData* _moduleData, llvm::Module* _module, const GTLCore::Type* _imageType, const GTLCore::Type* _type );
      /**
       * Generate a setAlpha function.
       */
      static llvm::Function* generatePixelSetAlpha( GTLCore::ModuleData* _moduleData, llvm::Module* _module, const GTLCore::Type* _pixelType, int channelsNb);
      /**
       * Generate a alpha function.
       */
      static llvm::Function* generatePixelAlpha( GTLCore::ModuleData* _moduleData, llvm::Module* _module, const GTLCore::Type* _pixelType, int channelsNb);
      /**
       * @return a pointer to the data pointer of a pixel (as a pointer to a vector of float)
       */
      static llvm::Value* accessPixelDataPtr( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _pixel);
      /**
       * @return a pointer to the data pointer of a pixel cast to a (float*)
       */
      static llvm::Value* accessPixelDataAsF32Ptr( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _pixel);
    private:
      static llvm::Value* accessPixelAlphaPos( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _pixel);
      static llvm::Function* createMemCpyFunction( llvm::Module* _module );
      static llvm::Value* accessPixelDataAsU8Ptr( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _pixel);
      static llvm::Value* callMemcpy( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _dst, llvm::Value* _src, llvm::Value* _n );
      static void setPixelCoordinates( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _dst, llvm::Value* _iVal, const GTLCore::Type* _iValType, llvm::Value* _jVal, const GTLCore::Type* _jValType );
      static void setPixelAlphaPos( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _dst, llvm::Value* _image, const GTLCore::Type* _imageType );
      /**
       * Call a "virtual" member of a structure, aka, a member with a pointer.
       */
      static llvm::Value* callVirtualMember( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, llvm::Value* _pointer, int _member_index , std::vector<llvm::Value*> arguments);
      /**
       * Call the ImageWrap data
       * @param callConst will call the const function, in case you just want to read data
       */
      static llvm::Value* callImageWrapData( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, const GTLCore::Type* _imageType, llvm::Value* _imageWrap, llvm::Value* _x, llvm::Value* _y, bool callConst);
      static llvm::Value* callImageAlphaPos( LLVMBackend::GenerationContext& _gc, llvm::BasicBlock* _currentBlock, const GTLCore::Type* _imageType, llvm::Value* _imageWrap);
    private:
      static int s_evaluatePixelsId;
  };
}

#endif
